#include "adb.h"
#include "sercommon.h"
#include "globals.h"

#ifndef	__GNUC__

#include "timer.h"
#include "panel.h"
#include "uart.h"

#else

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>

int	uart_handle;

#endif	/* __GNUC__ */

static byte adb_chksum;
static byte adb_lasterr;

/*
 * chyby
 * 1	Timeout
 * 2	CheckSum
 * 3	Length
 * 4	Unexpected data
 * 5	A/D board returned error
 * 8	A/D batteries low
 * 9	Other
 * vraci err
 */

/*
 * Doslo k chybe, na strane krabice se pokusim o restart sopjeni
 * (zacnu vysilat od zacatku), ale nejdriv necham prevodnik at si
 * dovyzblepta svoje data a dam jeste dve sekundy na vzpamatovani
 */
byte
adb_err(byte err)
{
	if (!err)
		return 0;
	adb_lasterr = err;
#ifndef __GNUC__
	p_mvputs(0, 3, "Chyba kom. s A/D:"); /* TODO: prelozit do cestiny */
	p_putidf(err, 2);
	jsleep(2 * JIFRATE);
#else
	printf("Chyba komunikace s A/D: %i\n", err);
	usleep(ADB_TIMEOUT * 1000000UL);
#endif
	return err;
}

byte
adb_geterr(void)
{
	register byte err = adb_lasterr;
	adb_lasterr = 0;
	return err;
}

/*
 * vysle len bajtu z data a POCKA az se uvolni vystupni buffer.
 * pokud len = 0 nebo data = NULL, jen pocka na uvolneni
 * vyst. buf. 
 * 
 * adb_out() a adb_in() by mely byt jedine dve platforme
 * zavisle fce, zbytek je snad prenositelny.
 */
void
adb_out(const void *data, byte len)
{
#ifndef __GNUC__
	if (data && len)
		ser_out(ADB_COM, data, len);
	while (ser_ostat(ADB_COM))
		go_sleep();
#else
	int ret;
	const byte *p = data;
	
	while (len) {
		ret = write(uart_handle, p, len);
		if (ret == -1) {
			perror("write");
			return;
		}
		len -= ret;
		p += ret;
	}
#endif /* __GNUC__ */
}

/*
 * precte len bajtu do data, pokud dojde k timeoutu, vola
 * adb_err(1) a vrati 1, jinak vrati 0.
 * 
 * adb_out() a adb_in() by mely byt jedine dve platforme
 * zavisle fce, zbytek je snad prenositelny.
 */
byte
adb_in(void *data, byte len)
{
#ifndef __GNUC__
	register dword timer = jiffies + ADB_TIMEOUT;
	
	ser_in(ADB_COM, data, len);
	while(ser_istat(ADB_COM)) {
		if (jiffies > timer)
			return 1;
		go_sleep();
	}
	return 0;
#else
	fd_set	set;
	struct timeval timeout;
	int ret;
	byte *p = data;

	while(len) {
		FD_ZERO(&set);
		FD_SET(uart_handle, &set);
		timeout.tv_sec = ADB_TIMEOUT;
		timeout.tv_usec = 0;
		switch(select(FD_SETSIZE, &set, NULL, NULL, &timeout)) {
			case 0: adb_err(1);
				return 1; /* timeout */
			case -1: perror("select");
				 return 9; /* error */
			case 1: break; /* input ready */
		}
		ret = read(uart_handle, p, len);
		if (ret == -1) {
			perror("read");
			return 9;
		}
		len -= ret;
		p += ret;
	}
	return 0;
#endif /* __GNUC__ */
}
	
byte
adb_init(void)
{
#ifdef __GNUC__
	uart_handle = open(ADB_COM, O_RDWR | O_NONBLOCK, 0);
	if (uart_handle == -1) {
		perror("open");
		return 1;
	}
#endif
	return 0;
}

byte
adb_done(void)
{
#ifdef __GNUC__
	if (close(uart_handle) == -1) {
		perror("close");
		return 1;
	}
#endif
	return 0;
}

/* Pripravi se na prenos zpravy do desky prvodniku (adb ~ a/d board) 
 * Predpoklada se ze port jiz byl inicializovan! */
void
adb_outstart(void)
{
	adb_chksum = 0;
	adb_out(NULL, 0); /* pockame az se uvolni vyst. buffer */
}

/* Vysle string tak jak je do adb */
void
adb_outs(const char *s)
{
	register byte x;

	for (x = 0; s[x]; adb_chksum += s[x++]);
	adb_out(s, x);
}

/* vysle byte v hexa do adb */
void
adb_outb(byte x)
{
	char buf[2];

	byte2ascii(buf, x);
	adb_chksum += buf[0] + buf[1];
	adb_out(buf, 2);
}

/* vysle int v hexa do adb */
void
adb_outint(int x)
{
	adb_outb(x >> 8);
	adb_outb(x & 0xFF);
}

/* Zakonci vysilani zpravy (vysle checksum a 0x0D) */
void
adb_outfinish(void)
{
	const char Terminator[] = {0x0D, 0};
	adb_outb(adb_chksum);
	adb_outs(Terminator);  /* Arnold strikes back! */
}


/* Pripravi se na prenos zpravy z desky prevodniku.  
 * Predpoklada se ze port jiz byl inicializovan! 
 * vraci 0: OK, 1: timeout */
#define adb_instart	adb_outstart

/* Precte z adb string delky len, tak jak je (zadne hexa) */
byte
adb_ins(byte *result, byte len)
{
	register byte err;
	
	if ((err = adb_in(result, len)))
		return err;
	while(len--)
		adb_chksum += *result++;
	return 0;
}

/* Precte cnt dat v hexa reprezentaci
 * (co jeden cnt to dva hexa znaky po seriaku) */
byte
adb_indata(void *result, byte cnt)
{
	char buf[2];
	register byte *p = result;
	register byte err;
	
	while(cnt--) {
		if ((err = adb_in(buf, 2)))
			return err;
		adb_chksum += buf[0] + buf[1];
		*p++ = ascii2byte(buf);
	}
	return 0;
}

byte
adb_infinish(void)
{
	char buf[3];
	register byte err;

	if ((err = adb_in(buf, 3)))
		return err;
	if (buf[2] != 0x0D)
		return 3 ;
	if (adb_chksum != ascii2byte(buf))
		return 2 ;

	return 0;
}

/* Precte ze seriaku byte v hexaformatu (2 znaky) */
byte
adb_inbyte(byte *result)
{
	register byte err;

	if ((err = adb_indata(result, 1)))
		return err;
	return 0;
}

/* Precte ze seriaku int (16tibitovy) v hexaformatu (4 znaky) */
byte
adb_inint(int *result)
{
	byte buf[2];
	register byte err;

	if ((err = adb_indata(buf, 2)))
		return err;
	*result = buf[0] << 8 | buf[1];
	return 0;
}
	
/* preskoci len bajtu prichazejicich z desky, updatuje pritom checksum. */
byte
adb_inskip(byte len)
{
	char buf;
	register byte err;

	for(; len; len--) {
		if ((err = adb_in(&buf, 1)))
			return err;
		adb_chksum += buf;
	}
	return 0;
}
#define TRYREAD(x) {if (adb_err(x)) goto restart;}

#ifdef ADB_READID 
byte
adb_rdid(byte addr, byte maxlen, byte *result)
{
	byte buf[2];
	register byte *p;
	register byte max;

restart:
	p = result;
	max = maxlen;
	
	adb_outstart();
	adb_outs("#C");
	adb_outb(addr);
	adb_outfinish();

	adb_instart();
	TRYREAD(adb_ins(buf, 1));
	if (buf[0] != '%') {
		adb_err(4);
		goto restart;
	}
	for(;;) {
		TRYREAD(adb_inbyte(p));
		if (!*p)
			break;
		if (--max)
			p++;
	}
	TRYREAD(adb_inbyte(buf));
	if (buf[0] != addr) {
		adb_err(9);
		return 1;
	}
	TRYREAD(adb_infinish());
	return 0;
}
#endif 

byte
adb_rdcalib(byte addr, byte chn, int *coef, int *off)
{
	byte buf;
restart:
	adb_outstart();
	adb_outs("#B");
	adb_outb(addr);
	adb_outb(chn);
	adb_outfinish();
	
	adb_instart();
	TRYREAD(adb_ins(&buf, 1));
	if (buf != '%') {
		adb_err(4);
		goto restart;
	}
	TRYREAD(adb_inint(coef));
	TRYREAD(adb_inint(off));
	TRYREAD(adb_infinish());
	return 0;
}

byte
adb_rdvals(byte addr, int *chnls, byte *cnt, byte *status)
{
	byte buf;
	byte numchnls;
	register byte i, read_cnt;
restart:
	adb_outstart();
	adb_outs("#D");
	adb_outb(addr);
	adb_outfinish();

	adb_instart();
	TRYREAD(adb_ins(&buf, 1));
	if (buf != '>') {
		adb_err(4);
		goto restart;
	}
	TRYREAD(adb_inbyte(status));
	TRYREAD(adb_inbyte(&numchnls));
	read_cnt = MIN(numchnls, *cnt);
	
	for(i = 0; i < read_cnt; i++)
		TRYREAD(adb_inint(chnls + i));
	TRYREAD(adb_inskip((numchnls - i) * 4));
	TRYREAD(adb_infinish());
	
	*cnt = read_cnt;
	return 0;
}
	
byte
adb_rddirect(byte addr, dword *chnls, byte *cnt, byte *status)
{
	byte buf[4];
	byte numchnls;
	register byte i, read_cnt;
	
restart:
	adb_outstart();
	adb_outs("#E");
	adb_outb(addr);
	adb_outfinish();

	adb_instart();
	TRYREAD(adb_ins(buf, 1));
	if (*buf != '>') {
		adb_err(4);
		goto restart;
	}
	TRYREAD(adb_inbyte(status));
	TRYREAD(adb_inbyte(&numchnls));
	read_cnt = MIN(numchnls, *cnt);
	
	for(i = 0; i < read_cnt; i++) {
		TRYREAD(adb_indata(buf, 4));
		chnls[i] = buf[1] << 16 | buf[2] << 8 | buf[3];
	}
	TRYREAD(adb_inskip((numchnls - i) * 8));
	TRYREAD(adb_infinish());
	
	*cnt = read_cnt;
	return 0;
}

byte
adb_wrcalib(byte addr, byte chnl, int koef, int offs)
{
	byte buf;
	
restart:
	adb_outstart();
	adb_outs("#A");
	adb_outb(addr);
	adb_outb(chnl);
	adb_outint(koef);
	adb_outint(offs);
	adb_outfinish();

	adb_instart();
	TRYREAD(adb_ins(&buf, 1));
	if (buf != '!') {
		adb_err(5);
		goto restart;
	}
	TRYREAD(adb_infinish());
	return 0;
}
