#include "sercom.h"
#include "globals.h"
#include "crc/crcmodel.h"
#include <stdarg.h>

#include "panel.h"  /* debug */
#include "num2str.h"  /* debug */

#ifndef __CKDM__
#include <string.h> /* kvuli printf */
#include <stdio.h>  /* kvuli strlen */
#endif

/* Changelog:

2006-09-14: sercom_init nyni nevraci byte, ale sercom_err. 
2006-09-14: Zverejneni sercom_in, sercom_out.
2006-09-14: sercom_init, sercom_done dostupne i ve verzi pro ckdm
2006-09-14: parametr baudrate u sercom_init
2006-09-14: platform specific funkce nahazeny do uart.c
2006-10-05: sercom_flushin nyni pouziva ser_purgebuf
*/

#define DEBUG

#include "timer.h"
#include "uart.h"

/*
 * vysle len bajtu z data a POCKA az se uvolni vystupni buffer.
 * pokud len = 0 nebo data = NULL, jen pocka na uvolneni
 * vyst. buf. 
 * 
 * sercom_out() a sercom_in() by mely byt jedine dve platforme
 * zavisle fce, zbytek je snad prenositelny.
 */
void
sercom_out(const void *data, size_t len)
{
	if (data && len)
		ser_out(SERCOM_COM, data, len);
	while (ser_olen(SERCOM_COM))
		go_sleep();
}

/*
 * Precte len bajtu do data. Ceka az SERCOM_TIMEOUT sekund nez data dorazi,
 * pak vraci err_timeout.
 */
enum sercom_err
sercom_in(void *data, size_t len)
{
	register uint32 timer = get_jiffies() + (JIFRATE * SERCOM_TIMEOUT);

	ser_in(SERCOM_COM, data, len);
	while(ser_istat(SERCOM_COM)) {
		if (get_jiffies() > timer) {
			return err_timeout;
		}
		go_sleep();
	}
	return err_ok;
}

void
sercom_flushin(void)
{
	ser_purgebuf(SERCOM_COM, SERCOM_TIMEOUT*JIFRATE);
	/*
	char buf;
	uint32 timer = get_jiffies() + (SERCOM_TIMEOUT * JIFRATE);
	ser_in(SERCOM_COM, &buf, 1);
	while (get_jiffies() < timer) {
		if (ser_ilen(SERCOM_COM)) {
			ser_in(SERCOM_COM, &buf, 1);
			timer = get_jiffies() + (SERCOM_TIMEOUT * JIFRATE);
		}
	}
	*/
}

/* Spolecne - sercom protokol *************************************************/

void
sercom_outdata(size_t len, push_data_func push, void *param)
{
	byte buf[3];
	cm_t crc = {16, 0x8005L, 0L, TRUE, TRUE, 0L};

	cm_ini(&crc);

	buf[0] = '#';
	buf[1] = len % 0x100;
	buf[2] = len / 0x100;
	sercom_out(buf, 3);
	while (len--) {
		buf[0] = push(param);
		cm_nxt(&crc, buf[0]);
		sercom_out(buf, 1);
	}
	buf[0] = cm_crc(&crc) % 0x100;
	buf[1] = cm_crc(&crc) / 0x100;
	buf[2] = 0x0a;
	sercom_out(buf, 3);
}

byte
sercom_indata(pop_data_func pop, void *param)
{
	byte buf[3];
	cm_t crc = {16, 0x8005L, 0L, TRUE, TRUE, 0L};
	enum sercom_err retval;
	unsigned int len;

	cm_ini(&crc);

	if ((retval = sercom_in(buf, 3)) != err_ok)
		goto err;
	if (buf[0] != '#') {
		retval = err_begin;
		goto err;
	}
	len = buf[1] + (buf[2] * 0x100);
	while(len--) {
		if ((retval = sercom_in(buf, 1)) != err_ok)
			goto err;
		cm_nxt(&crc, buf[0]);
		if ((retval = pop(param, buf[0])) != err_ok)
			goto err;
	}
	if ((retval = sercom_in(buf, 3)) != err_ok)
		goto err;
	if ((buf[0] != cm_crc(&crc) % 0x100) || (buf[1] != cm_crc(&crc) / 0x100)) {
		retval = err_chksum;
		goto err;
	}
	if (buf[2] != 0x0a) {
		retval = err_end;
		goto err;
	}
	return err_ok;

err:
#if (defined DEBUG && !defined __CKDM__)
	printf("%s\n", sercom_strerr(retval));
#endif 
	/* Pokud je timeout a jsem server, flushlo by se mi vsechno co bych mohl
	 * potrebovat, jako server totiz dlouho cekam na prikaz a tim casto
	 * timeoutuju. Pokud by doslo k timeoutu z jineho duvodu, v dalsim behu se
	 * objevi jina chyba (pravdepodobne err_begin) a vse je korektne osetreno.
	 */ 
	if (retval != err_timeout)
		sercom_flushin();
	return retval;
}

static byte
pushstr(const void *p)
{
	register char **s = (void *)p;
	register const byte retval = **s;
	(*s)++;
	return retval;
}

void
sercom_outstr(const char *s)
{
	sercom_outblck((const byte *)s, strlen(s) + 1);
}	

struct popstr_s {
	byte *buf;
	unsigned int len;
};

static enum sercom_err
popstr(void *p, byte ch)
{
	struct popstr_s *s = p;
	if (!s->len)
		return err_pop;
	s->len--;
	*(s->buf) = ch;
	s->buf++;
	return err_ok;
}

enum sercom_err
sercom_instr(char *buf, size_t maxlen)
{
	enum sercom_err retval;
	struct popstr_s s;
	s.buf = (byte *)buf;
	s.len = maxlen;
	retval = sercom_indata(popstr, &s);
	s.buf[maxlen - 1] = '\0';
	return retval;
}	

/**********************************/

void
sercom_outblck(const byte *data, size_t len)
{
	sercom_outdata(len, (byte(*)(void *))pushstr, &data);
}	

enum sercom_err
sercom_inblck(byte *data, size_t *len)
{
	enum sercom_err retval;
	struct popstr_s s;

	s.buf = data;
	s.len = *len;
	retval = sercom_indata(popstr, &s);
	*len = *len - s.len;
	return retval;
}	

void
uint32_str(byte *buf, uint32 num)
{
	buf[0] = (num >> 0 ) & 0xff;
	buf[1] = (num >> 8 ) & 0xff;
	buf[2] = (num >> 16) & 0xff;
	buf[3] = (num >> 24) & 0xff;
}

uint32
str_uint32(const byte *buf)
{
	/* !KOUZLO! */
	/* zakomentovany radek nefunguje, generuje "magickou" chybu.
	 * Pokud dam za vstup buf hodnoty buf[0] = buf[1] = buf[2] = buf[3] = 0xff,
	 * vypocita se podle prvniho radku 0xfffffeff. 
	 * Druhy radek funguje korektne (0xffffffff).
	 * opravdu NETUSIM proc :-)
	 */
	register uint32 retval;

	/* retval = buf[0] + (buf[1]*0x100) + (buf[2]*0x10000) + (buf[3]*0x1000000); */
	retval = (uint32)buf[0] |
		((uint32)buf[1] << (uint32)8) |
		((uint32)buf[2] << (uint32)16) |
		((uint32)buf[3] << (uint32)24);
	return retval;
}

size_t
sercom_putf(byte *buf, const char *format, ...)
{
	const char *f, *tmp;
	va_list arg;
	size_t len = 0;

	UNUSED_ARG(__builtin_va_alist);

	va_start(arg, format);
	for (f = format; *f; f++) {
		switch(*f) {
			case 'c': 
				*buf++ = va_arg(arg, uint32) % 0x100; 
				len++;
				break;
			case 'i': 
			case 'I': 
				uint32_str((byte *)buf, va_arg(arg, uint32));
				buf += 4;
				len += 4;
				break;
			case 's':
				tmp = va_arg(arg, const char *);
				while('\0' != (*buf++ = *tmp++))
					len++;
				len++;
				break;
		}
	}
	va_end(arg);
	return len;
}

int
sercom_outf(const char *format, ...)
{
	const char *f;
	va_list arg;
	int len = 0;
	byte buf[4];
	cm_t crc = {16, 0x8005L, 0L, TRUE, TRUE, 0L};
	const char *tmp;

	cm_ini(&crc);

	UNUSED_ARG(__builtin_va_alist);

	va_start(arg, format);
	for (f = format; *f; f++) {
		switch(*f) {
			case 'c': len += 1; if(va_arg(arg, uint32)); break;
			case 'i': case 'I': len += 4; if(va_arg(arg, uint32)); break;
			case 's': len += strlen(va_arg(arg, const char *)) + 1; break;
		}
	}
	va_end(arg);

	buf[0] = '#';
	buf[1] = len % 0x100;
	buf[2] = len / 0x100;
	sercom_out(buf, 3);

	va_start(arg, format);
	for (f = format; *f; f++) {
		switch(*f) {
			case 'c': 
				buf[0] = va_arg(arg, uint32) % 0x100; 
				cm_nxt(&crc, buf[0]);
				sercom_out(buf, 1); 
				break;
			case 'i': 
			case 'I': 
				uint32_str(buf, va_arg(arg, uint32));
				cm_blk(&crc, buf, 4);
				sercom_out(buf, 4);
				break;
			case 's':
				tmp = va_arg(arg, const char *);
				do {
					cm_nxt(&crc, *tmp);
					sercom_out(tmp, 1);
				} while(*tmp++);
				break;
		}
	}
	va_end(arg);
	
	buf[0] = cm_crc(&crc) % 0x100;
	buf[1] = cm_crc(&crc) / 0x100;
	buf[2] = 0x0a;
	sercom_out(buf, 3);
	return 0;
}

int
sercom_vscanf(const byte *data, const char *format, va_list arg)
{
	register const byte *p;
	const char *f;
	int retval = 0;
		/* Len je podporovano, ale zatim ho neznam, tohle
		 * je kdyby do implementace prisel dalsi parametr
		 * urcujici delku dat. */
	unsigned int len = 0x7FFF; 
	byte assign = 1;

	p = data;
	for (f = format; *f; f++) {
		if (*f == '.') {
			assign = 0;
			continue;
		}	
		switch(*f) {
			case 'c': 
				if (p + 1 > data + len)
					goto finish;
				if (assign)
					*(va_arg(arg, uint32*)) = *p;
				p += 1;
				break;
			case 'i':
			case 'I':
				if (p + 4 > data + len)
					goto finish;
				if (assign)
					*(va_arg(arg, uint32*)) = str_uint32(p);
				p += 4;
				break;
			case 's':
				if (assign)
					*(va_arg(arg, const byte **)) = p;
				do {
					if (p + 1 > data + len)
						goto finish;
				} while(*p++);
				break;
		}
		assign = 1;
		retval += 1;
	}
finish:
	va_end(arg);
	return retval;
}


int
sercom_scanf(const byte *data, const char *format, ...)
{
	va_list arg;

	UNUSED_ARG(__builtin_va_alist);

	va_start(arg, format);
	return sercom_vscanf(data, format, arg);
}


const char *
sercom_strerr(enum sercom_err error)
{
	switch(error) {
		case err_ok:		return "OK";
		case err_timeout:	return "Timeout";
		case err_begin:		return "Begin";
		case err_chksum:	return "ChkSum";
		case err_unknown:	return "Unknown";
		case err_pop:		return "Pop";
		case err_end:		return "End";
		default:		return "Invalid(SNH)";
	}
}
