/* 
 * sercom.h
 * Protokol po seriovou binarni komunikaci s checksumem.
 * jeden packet vypada takto:
 * # LoLen HiLen Data LoCRC HiCRC 0x0A
 * kde 
 * # je znak #
 * Lolen je nizky bajt delky
 * HiLen je vysoky bajt delky
 * Data je Len po sobe jdoucich bajtu
 * LoCRC je nizky bajt CRC souctu
 * LoCRC je vysoky bajt CRC souctu
 * 0x0A je znak konce radky.
 * 
 */

#ifndef _SERCOM_H_
#define _SERCOM_H_

#include "globals.h"
#include "uart.h"
#include <stdarg.h>

/* Ma se zakompilovat sercom_rdid() ? */
/* #define SERCOM_READID  */

/* na kterem UARTu budeme komunikovat ? */
#define SERCOM_COM	1

/* Timeout prijmu zprav (kdyz vyprsi, vypise se varovna hlaska na displej
 * a zkousi se to donekonecna znovu) */
#ifndef __GNUC__
#define SERCOM_TIMEOUT	2	/* [seconds] */
#else
extern int SERCOM_TIMEOUT;		/* [seconds] */
#endif
/* Debughlasky ? Pise jaka chyba na kterem radku nastala. */
#define SERCOM_DEBUG

/* Low level funkce ***********************************************************/

/* Navratove hodnoty nekterych funkci. */
enum sercom_err {
	err_ok, /* Vse OK */
	err_timeout, /* Timeout */
	err_begin, /* Chybny pocatecni znak */
	err_chksum, /* chybny checksum */
	err_unknown, /* jina chyba */
	err_pop, /* chyba popovaci funkce (viz sercom_indata()) */
	err_end /* chybny ukoncovaci znak */
};

typedef byte (* push_data_func)(void *);
typedef enum sercom_err (* pop_data_func)(void *, byte);

/* Vraci textovou reprezentaci chyby, pokud neni
 * chyba zavedena vraci "Invalid"
 */
const char *
sercom_strerr(enum sercom_err error);
	
/* Parametr pro sercom_init, v fifo_* pripadech se komunikuje
 * obousmerne pres ddve fifo, v tty* pripadech uz primo
 * pres dany seriak. 
 * Windows: ttys0 ~ com1, ttys1 ~ com2
 * Linux: Primo odpovida prislusnemu zarizeni ttySx
 * CKDM: ttys0 ~ Async0, ttys1 ~ Async1
 */
 
enum sercom_inittype {
	ttys0, ttys1, fifo_server, fifo_tester
};

/* Inicializuje vse potrebne pro ser. prenos, zvol
 * jednu z hodnot fifo_server, fifo_tester, ttys0, ttys1
 * Vraci 0 pri OK, 1 pri chybe.
 */
enum sercom_err
sercom_init(enum sercom_inittype, int baudrate);

/* Uzavre handly, vraci 0 pri OK, 1 pri chybe */
enum sercom_err
sercom_done(void);

/* Funkce pro sercom protokol *************************************************/

/* TODO dokumentace */
byte
sercom_indata(pop_data_func pop, void *param);

/* TODO dokumentace */
void
sercom_outdata(size_t len, push_data_func push, void *param);
	
/* Vysle na seriak string ve formatu danem protokolem.
 */
void
sercom_outstr(const char *s);

/* Pokusi se ze seriaku prijmout string dany protokolem.
 * maxlen je maximalni pocet bajtu ktery se ulozi do buf,
 * v pripade prekroceni fce vraci err_pop. buf[maxlen - 1]
 * je po ukonceni funkce vzdy 0 (korektni c-string).*/
enum sercom_err
sercom_instr(char *buf, size_t maxlen);

/* Vysle na seriak buffer data delky len ve formatu danem protokolem/
 */
void
sercom_outblck(const byte *data, size_t len);

/* Pokusi se ze seriaku prijmout buffer o delce az *len. V *len
 * je po navratu z fce skutecna delka prijatych dat. V pripade
 * ze dat bylo vice nez byl buffer schopen pojmout fce vraci err_pop.
 */
enum sercom_err
sercom_inblck(byte *data, size_t *len);

/* ctyrbajtovy cislo na string */
void
uint32_str(byte *buf, uint32 num);

/* String na ctyrbajtovy cislo */
uint32
str_uint32(const byte *buf);

/*
 *
 * Posle na seriak data podle formatu.
 * format:
 * s .. string (null terminated) (typ const char *)
 * c .. integer 0..255 (typ uint32)
 * i .. integer 0 .. 2^32-1 (typ uint32)
 * I .. integer -2^31 .. 2^31 - 1 (typ int32)
 *
 * Priklad:
 * sercom_outf("sii", "Ahoj svete", 0x666, 666)
 * vysle na seriak string ahoj svete a dva uint32 0x666 a 666
 */
int
sercom_outf(const char *format, ...);

/*
 * To same co sercom_outf, ale do bufferu. Vraci delku dat do bufferu
 * zapsanych.
 */
size_t
sercom_putf(byte *buf, const char *format, ...);

/*
 * Rozparsuje data, ktera byla vyslana pomoci sercom_outf
 * a prijata pomoci sercom_indata.
 * format: "i.cs"
 * i .. uint32* , ctyri bajty (0 .. 2^32 - 1)
 * I ..  int32* , ctyri bajty (-2^31 .. 2^31 - 1)
 * c .. uint32* , jeden bajt (0 .. 255)
 * s .. char**, nulou ukonceny retezec, parametr bude ukazovat nekam
 * do *data.
 * Pokud je nektery parametr predejit teckou ('.'), znamena
 * to ze se sice ocekava parametr daneho typu, ale nema se nikam
 * ukladat (v ... pro nej neni vyhrazene misto).
 * Priklad: sercom_scanf("s.ii",  &str, &end)
 * Navratova hodnota: Pocet uspesne preparsovanych parametru.
 */
int
sercom_scanf(const byte *data, const char *format, ...);

/* 
 * Jako sercom_scanf, ale misto ... bere va_list.
 * Na konci sam provede va_end(arg).
 */
int
sercom_vscanf(const byte *data, const char *format, va_list arg);

#endif /* _SERCOM_H_ */
