#include <string.h>
#include "table.h"
#include "globals.h"
#include "stdarg.h"
#include "gpib.h"
#include "panel.h"
#include "misc.h"
#include "atol.h"
#include "parse_float.h"
#include "timer.h"
#include "buzzer.h"
#include "ui.h"

/* Changelog:
 * 2006-10-11: Includuju <string.h> misto "block.h"
 */
static uint32 table_flags;

static enum table_state {
	table_idle,
	table_referencing
} table_state;

/* Pokud je zadefinovano, stul se pouze jednoduse simuluje.  Simuluji nekonecne
 * rychly stul, kazdy presun je proveden okamzite.  */

/* #define SIMULATION */

/* SIMULACE ****************************************************************/

#ifdef SIMULATION

static int32 table_x, table_y;	/* poloha vztazena k referencni poloze */
static int32 table_zero_x, table_zero_y; /* hodnota citace v referencni poloze */

enum cmd_err
table_init(void)
{
	table_x = table_y = table_zero_x = table_zero_y = 0;
	table_flags = 0;
	table_set_local(true);
	table_flags |= TABLE_BIT_ALIVE;
	return cmd_ok;
}

enum cmd_err
table_get_id(char *str, size_t maxlen)
{
	strncpyx(str, "Simulace otocneho stolu", maxlen);
	return cmd_ok;
}

enum cmd_err
table_get_flags(uint32 *flags)
{
	*flags = table_flags;
	return cmd_ok;
}
	
enum cmd_err
table_goto(int32 x, int32 y)
{
	table_x = x + table_zero_x;
	table_y = y + table_zero_y;
	return cmd_ok;
}


enum cmd_err
table_go_reference(void)
{
	table_flags |= TABLE_BIT_REFERENCED;
	table_goto(table_zero_x, table_zero_y);
	return cmd_ok;
}

enum cmd_err
table_set_zero(int32 x, int32 y)
{
	table_zero_x = x; table_zero_y = y;
	return cmd_ok;
}

enum cmd_err
table_go_zero(void)
{
	return table_goto(0, 0);
}

enum cmd_err
table_get_zero(int32 *x, int32 *y)
{
	*x = table_zero_x;
	*y = table_zero_y;
	return cmd_ok;
}

enum cmd_err
table_kill(void)
{
	return cmd_ok;
}

enum cmd_err
table_get_pos(int32 *x, int32 *y)
{
	*x = table_x - table_zero_x;
	*y = table_y - table_zero_y;
	return cmd_ok;
}
	
enum cmd_err
table_set_local(bool local)
{
	if (local) {
		table_flags |= TABLE_BIT_LOCAL;
	} else {
		table_flags &=~ TABLE_BIT_LOCAL;
	}
	return cmd_ok;
}

enum cmd_err
table_invert(uint32 invert)
{
	UNUSED_ARG(invert);
	return cmd_ok;
}

enum cmd_err
table_geterr(uint32 *err)
{
	*err = 0;
	return cmd_ok;
}

enum cmd_err
table_is_busy(uint32 *busy)
{
	UNUSED_ARG(busy);
	return cmd_ok;
}		
#else 

/* Skutecne HW **************************************************************/

#define GPIB_PRE \
	byte gpib_tries; \
	for (gpib_tries = 0; gpib_tries < 3; gpib_tries++) { \
		if (gpib_get_status() != GPIB_NONE) \
			gpib_reset();	

#define GPIB_POST \
		if (gpib_get_status() == GPIB_NONE) { \
			table_flags |= TABLE_BIT_ALIVE; \
			return cmd_ok; \
		} \
	}\
	table_flags &=~ TABLE_BIT_ALIVE; \
	return cmd_comm;

/* k retezci buf pripoji " c(-)xxx.xxx", 
 * kde (-)xxx.xxx je (zaporne)num/1000 a c je parametr c.
 */
static void
table_appendmsg(char *buf, char c, int32 num)
{
	while(*buf++);
	buf[-1] = ' ';
	*buf++ = c;
	if (num < 0) {
		*buf++ = '-';
		num = 0 - num;
	}
	s_putid0(buf, num/1000, 3);
	buf[3] = '.';
	s_putid0z(buf + 4, num, 4);
}


static enum cmd_err
table_wrstr(const char *buf)
{
	GPIB_PRE
	gpib_wrstr(TABLE_ADDR, buf);
	GPIB_POST
}

static enum cmd_err
table_send(const char *msg, byte n_params, ... /*char c1, int32 p1, ... */)
{
	char buf[30];
	va_list arg;
	char c;
	int32 p;
	
	UNUSED_ARG(__builtin_va_alist);
	va_start(arg, n_params);
	
	strcpy(buf, msg);
	while(n_params--) {
		c = va_arg(arg, char);
		p = va_arg(arg, int32);
		table_appendmsg(buf, c, p);
	}
	va_end(arg);
	return table_wrstr(buf);
}

static enum cmd_err
table_recv(char *buf, size_t maxlen)
{
	byte cnt;

	GPIB_PRE

	cnt = gpib_read(TABLE_ADDR, buf, maxlen);
	buf[cnt] = 0;

	GPIB_POST
}


/* vraci true, pokud se povedlo, jinak false */
static bool
scancoord(const char *s, int32 *c)
{
	float f;
	
	if (!parse_float(s, &f))
		return false;
	*c = (int32)(f*1000.0);
	return true;
}

/* vraci true kdyz se povedlo. 

Format souradnic je pekne hnusnej. Stul mi jako jednu souradnici
muze poslat cokoliv z napr.

0
1
.12
-.12
1.12
-1.12345 (!, je treba zaokrouhlit na petitisiciny stupne, zde orezavam na 1000iny)
5E-03
-5E-03 (aspon to vypada ze E je vzdycky velky)
4.98902E-03 (fuj!)

nevypada to ze by se vyskytovaly kladny exponenty.
souradnice muze byt (ale nemusi) tesne za oznacenim, napr.

X 1
X-1

Kazdopadne je to HUMUS! do pekla s tvurcem :-)
*/

static bool
table_scancoords(const char *s, int32 *x, int32 *y)
{
	 /* X */
	while (*s && *s != 'X')
		s++;
	if (!*s)
		return false;
	s++;
	if (!scancoord(s, x))
		return false;
	 /* Y */
	while (*s && *s != 'Y')
		s++;
	if (!*s)
		return false;
	s++;
	if (!scancoord(s, y))
		return false;
		
	return true;
}

/* Funkce podle protokolu **************************************************/

enum cmd_err
table_init(void)
{
	GPIB_PRE

	table_flags = 0;

	GPIB_POST
}

/* Nyni testuji busy tak, ze zkusim prazdny prikaz ("   ") Pointa je v tom, ze
 * kdyz stul neco provadi (tj. je busy), a pokusim se mu poslat nejakou zpravu,
 * prijme alespon jeden bajt (!), ale dalsi bajty uz ne (neustale drzi NRFD =
 * T).  Pokud busy neni, zpravu v pohode prijme, ale posilam mu jenom mezery
 * ktere vlastne nic nedelaji. Mohl bych na testovani udelat spec. funkci,
 * ale proc kdyz muzu pouzit gpib_wrstr...
 *
 * Jeden z dusledku je, ze testovani trva az GPIB_TIMEOUT, ale to snad nicemu
 * nevadi..
 */
enum cmd_err
table_get_flags(uint32 *flags)
{
	enum cmd_err err;
	size_t tmp;

	table_flags |= TABLE_BIT_ALIVE;
	table_flags &=~ TABLE_BIT_BUSY;
	*flags = table_flags;

	if (gpib_get_status() != GPIB_NONE)
		gpib_reset();

		/* Zjistim zda je nazivu a busy */
       	tmp = gpib_wrstr(TABLE_ADDR, "   ");
	if (gpib_get_status() != GPIB_NONE) {
		if (tmp == 0 && (gpib_get_status() & GPIB_NOT_ALIVE)) { /* Vubec nezije */
			table_flags &=~ TABLE_BIT_REFERENCED;
			table_flags &=~ TABLE_BIT_ALIVE;
			table_state = table_idle;
		} else { /* zije, ale je busy */
			table_flags |= TABLE_BIT_ALIVE;
			table_flags |= TABLE_BIT_BUSY;
		}
		gpib_ifc();
		gpib_reset();
		err = cmd_ok;
		goto finish;
	}
	err = cmd_ok;

	table_flags &=~ TABLE_BIT_BUSY;
	table_flags |= TABLE_BIT_ALIVE;
	if (table_state == table_referencing)
		table_flags |= TABLE_BIT_REFERENCED;
	table_state = table_idle;

finish:
	*flags = table_flags;
	return err;
}

enum cmd_err
table_get_last_error(uint32 *last_error)
{
	enum cmd_err err;
	char errbuf[5];
	byte i;

	/* Chyby majici za nasledek ztratu polohy (odreferencovani) */
	const byte unreferencing_errors[] = {
		TABLE_ERR_POWER,
		TABLE_ERR_MOTORS,
		TABLE_ERR_POS,
		TABLE_ERR_STOP,
		};

	if (cmd_ok != (err = table_query("M93", errbuf, sizeof(errbuf))))
		return err;

	/* Zjistim posledni chybu stolu, pripadne zrusim flag referenced */
	*last_error = atol(errbuf);
	for (i = 0; i < sizeof(unreferencing_errors)/sizeof(byte); i++)
		if (*last_error == unreferencing_errors[i])
			table_flags &=~ TABLE_BIT_REFERENCED;

	return cmd_ok;
}

static enum cmd_err
pre_cmd(void)
{
	uint32 flags;
	enum cmd_err err;

	if (cmd_ok != (err = table_get_flags(&flags)))
		return err;
	if (!(flags & TABLE_BIT_ALIVE))
		return cmd_comm;
	if (flags & TABLE_BIT_BUSY)
		return cmd_wait;
	return cmd_ok;
}

#define PRE_CMD \
	enum cmd_err err; \
	if (cmd_ok != (err = pre_cmd())) \
		return err;
	
static enum cmd_err
table_gtl(void)
{
	GPIB_PRE
	gpib_gtl(TABLE_ADDR);
	GPIB_POST
}

/* Pokud je local true, rameno jde do lokalniho rezimu, jinak se lok. ovladani
 * zakaze.
 * TODO nastavovat local bit ve vsech prikazech? nebo se na nej vysrat uplne?
 */
enum cmd_err
table_set_local(bool local)
{
	uint32 tmp;

	PRE_CMD

	if (local) {
		if (
			cmd_ok != (err = table_send("G50", 0)) ||
			cmd_ok != (err = table_gtl())
		   )	
			return err;
		table_flags |= TABLE_BIT_LOCAL;
	} else {
		if (cmd_ok != (err = table_get_flags(&tmp)))
			return err;
		table_flags &=~ TABLE_BIT_LOCAL;
	}
	return cmd_ok;
}

enum cmd_err
table_get_id(char *str, size_t maxlen)
{
	PRE_CMD

	return table_query("*IDN?", str, maxlen);
}

enum cmd_err
table_goto(int32 x, int32 y)
{
	PRE_CMD

	if (
		cmd_ok != (err = table_wrstr("G90")) ||
		cmd_ok != (err = table_send("G01", 2, 'X', x, 'Y', y))
	   );
	return err;
}

enum cmd_err
table_go_reference(void)
{
	PRE_CMD

	table_state = table_referencing;
	return table_wrstr("G98");
}

enum cmd_err
table_set_zero(int32 x, int32 y)
{
	PRE_CMD

	if (
		cmd_ok != (err = table_send("G92", 2, 'X' ,x, 'Y', y)) ||
		cmd_ok != (err = table_send("G99", 2, 'X', x, 'Y', y))
	   );
	return err;
}

enum cmd_err
table_get_zero(int32 *x, int32 *y)
{
	char buf[30];

	PRE_CMD

	if (cmd_ok != (err = table_query("M92", buf, sizeof(buf))))
		return err;
	return table_scancoords(buf, x, y) ? cmd_ok : cmd_invalid;
}

enum cmd_err
table_get_pos(int32 *x, int32 *y)
{
	char buf[30];

	PRE_CMD

	if (cmd_ok != (err = table_query("M25", buf, sizeof(buf))))
		return err;
	return table_scancoords(buf, x, y) ? cmd_ok : cmd_invalid;
}

enum cmd_err
table_invert(uint32 invert)
{
	PRE_CMD

	return table_wrstr(invert ? "G61" : "G60");
}

enum cmd_err
table_kill()
{
	p_msg("Kill neimplementovan!");
	go_stop(true);
	return cmd_ok;
}

/* Nedokumentovano */
enum cmd_err
table_query(const char *q, char *buf, int maxlen)
{
	enum cmd_err err;

	if (cmd_ok != (err = table_wrstr(q))) {
		return err;
	}
	return table_recv(buf, maxlen);
}
	
#endif /* !SIMULATION */
