/*
#define SHOW_AD_VALUE
#define TEST_INPUTS
*/

#include "globals.h"
#include "ui.h"
#include "arm.h"
#include "panel.h"
#include "p_num_x_x.h"
#include "timer.h"
#include "uart.h"
#include "sercom.h"
#include "buzzer.h"
#include "block.h"
#include "luxmetr.h"
#include "gpib.h"
#include "table.h"
#include "ad.h"

/* Jen konstanty pro znaky se spec. symboly */
#define P_DEG 1
#define P_HOURGLASS 2
#define P_OK 3
#define P_BAD 'x'

/* Max. delka prikazu prichoziho po seriaku */
#define MAXCMD 30

struct arm_program arm_programs[4] = {{0}, {0}, {0}, {0}};

const char *
cmd_strerr(enum cmd_err err)
{
	const char * const err_strs[] = {
		"ok", 
		"terminated", 
		"timeout", 
		"invalid", 
		"comm", 
		"wait", 
		"local", 
		"stop_pressed", 
		"illegal_movement"
	};
	return err_strs[err];
}

static void
puterror(const char *domain, const char *err)
{
	p_clearrow(2); 
	if (!domain && !err)
		return;
	p_mvputs(0, 2, domain); 
	p_puts(": "); 
	p_puts(err);
}

/* Struktura obrazovky je nasledujici:
"+xxx.xxxD  ?    PC H"
"Rameno mimo rozsah  "
"Ram: invalid        "
"F1: Vyrobce         "

Na prvnim radku je poloha. D je znacka stupnu, ? je vysvicen pokud neni znama
poloha ramene (nebylo referencovano nebo vypadlo z vazby), PC je vysviceno
pokud je rameno nyni ovladano z PC (tj. panel je zamceny), H je znacka hodin
pokud se rameno prave pohybuje nebo znacka OK pokud stoji.

Na druhem radku se ukazuji informace dulezite pro stav ramene (rameno seplo
koncovy spinac, ...)

na tretim radku jsou chyby komunikace s PC a s radicem motoru.

*/

static void
sercmds_init(void)
{
	ser_init(SERCOM_COM, 38400, 0);
	ser_flushbuf(SERCOM_COM);
}

static int
sercom_outint(uint32 x)
{
	return sercom_outf("i", x);
}
		
#define EQ(a, b) (strncmp((char *)(a), (char *)(b), sizeof(b) - 1) == 0)

static bool
lux_callback(void)
{
	byte cmd[MAXCMD];
	unsigned int cmdlen;
	enum sercom_err err;
	uint32 flags;
	
	if (ser_isbufempty(SERCOM_COM))
		return false;

	cmdlen = sizeof(cmd);
	if ((err = sercom_inblck(cmd, &cmdlen)) != err_ok)
		return false;
	if (EQ(cmd, "L_Flags")) {
		err = lux_get_flags(&flags);
		sercom_outf("ii", (uint32)err, flags); 
		return false;
	} else if (EQ(cmd, "L_Get")) {
			/* cekame na odmer */
		sercom_outf("ii", (uint32)cmd_wait, (uint32)0);	
		return false;
	} else if (EQ(cmd, "L_Stop")) {
		sercom_outint(cmd_ok);
		return true;
	} else {
		sercom_outint(cmd_invalid);
	}
	return false; /* ostatni ignoruji */
}

static void
gpib_try_reset(void)
{
	if (gpib_get_status() != GPIB_NONE)
		gpib_reset();
}

uint32
gpib_status_as_uint32(void)
{
	return (uint32)(gpib_get_status() != GPIB_NONE ? cmd_comm : cmd_ok);
}

enum cmd_err
stop_button_pressed(void)
{
	/* Jakmile jednou stisknuta, necht drzi az do resetu, stopka je zavazna situace */
	static bool stop_pressed = false;

	/*
	if (IO_PDR6.bit.D67)
		stop_pressed = true
	*/
	if (stop_pressed)
		p_putrow(1, "Stisknuto STOP!");
	return stop_pressed;
}

/* Vraci true, pokud prisel nejaky byte po seriaku, false pokud ne. */
static bool
sercmds_process(void)
{
	byte cmd[MAXCMD];
	unsigned int cmdlen;
	enum sercom_err err;
	enum cmd_err cerr;
	uint32 flags;
	int32 lux_val;
	
	if (ser_isbufempty(SERCOM_COM))
		return false;

	cmdlen = sizeof(cmd);
	if ((err = sercom_inblck(cmd, &cmdlen)) != err_ok) {
		if (err == err_timeout)
			return false;
		puterror("PC", sercom_strerr(err));
		return true;
	}
	puterror(NULL, NULL);

	if (stop_button_pressed()) {
		sercom_outint((uint32)cmd_stop_pressed);
		return true;
	}

	switch(cmd[0]) {
		case 'C': goto console;
		case 'L': goto luxmeter;
		case 'A': goto arm;
		case 'T': goto table;
		default: goto unknown;
	}

console:
	if (EQ(cmd, "C_ID")) {
		sercom_outf("is", (uint32)cmd_ok, NAPALM_TEXT);
	} else if (EQ(cmd, "C_Send")) {
		char *data;
		uint32 who;
		sercom_scanf(cmd, ".sis", &who, &data);
		gpib_try_reset();
		gpib_wrstr(who, data);
		sercom_outint(gpib_status_as_uint32());
	} else if (EQ(cmd, "C_Recv")) {
		char data[MAXCMD];
		uint32 who;
		byte cnt;

		sercom_scanf(cmd, ".si", &who);
		gpib_try_reset();
		cnt = gpib_read(who, data, sizeof(data) - 1);
		data[cnt] = '\0';
		sercom_outf("is", gpib_status_as_uint32(), data);
	/* Nedokumentovano: GPIB Interface clear */
	} else if (EQ(cmd, "C_IFC")) {
		gpib_ifc();
		sercom_outint(gpib_status_as_uint32());
	} else goto unknown;
	return true;

luxmeter:
	if (EQ(cmd, "L_Reset")) {
		sercom_outint(lux_init()); 
	} else if (EQ(cmd, "L_Flags")) {
		err = lux_get_flags(&flags);
		sercom_outf("ii", (uint32)err, flags); 
	} else if (EQ(cmd, "L_RDis")) {
		sercom_outint((uint32)cmd_ok);
		/* Osetreno nize */
	} else if (EQ(cmd, "L_Start")) {
		lux_get_flags(&flags);
		if (flags & LUX_BIT_MEASURING) {
			sercom_outint(cmd_wait);
		} else {
			sercom_outint(cmd_ok);
			lux_measure(lux_callback);
		}
	} else if (EQ(cmd, "L_Stop")) {
		/* Nemeri se, Stop nema smysl */
		sercom_outint(cmd_invalid); 
	} else if (EQ(cmd, "L_Get")) {
		err = lux_get_val(&lux_val);
		sercom_outf("iI", (uint32)err, lux_val);
	/* Nedokumentovano: Precte displej luxmetru*/
	} else if (EQ(cmd, "L_Q")) {
		char *query;
		char result[30];

		cerr = lux_rddata(result, sizeof(result));
		sercom_outf("is", (uint32)cerr, result);
	/* Nedokumentovano: Nastavi rozsah luxmetru */
	} else if (EQ(cmd, "L_R")) {
		uint32 range;
		sercom_scanf(cmd, ".si", &range);
		sercom_outf("i", (uint32)(lux_set_range(range)));
	} else 
		goto unknown;

	lux_set_local(EQ(cmd, "L_RDis"));
	return true; 
	
arm:
	if (EQ(cmd, "A_ID")) {
		sercom_outf("is", (uint32)cmd_ok, "Rameno " ARM_FIRMWARE_VERSION);
	} else if (EQ(cmd, "A_Flags")) {
		cerr = arm_get_flags(&flags);
		sercom_outf("ii", (uint32)cerr, flags);
	} else if (EQ(cmd, "A_RDis")) {
		sercom_outint((uint32)cmd_ok);
		/* Osetreno nize */
	} else if (EQ(cmd, "A_Goto")) {
		int32 where;
		sercom_scanf(cmd, ".si", &where);
		if (cmd_ok != (err = arm_set_movement_params(&arm_movement_full)))
			sercom_outint(err);
		sercom_outint(arm_goto(where));
	} else if (EQ(cmd, "A_Kill")) {
		sercom_outint(arm_kill());
	} else if (EQ(cmd, "A_Pos")) {
		int32 pos;
		cerr = arm_get_pos(&pos);
		sercom_outf("iI", (uint32)cerr, pos);
	} else if (EQ(cmd, "A_SetZero")) {
		int32 zero;
		sercom_scanf(cmd, ".si", &zero);
		sercom_outint(arm_set_zero(zero));
	} else if (EQ(cmd, "A_GetZero")) {
		int32 zero;
		cerr = arm_get_zero(&zero);
		sercom_outf("iI", (uint32)cerr, zero);
	} else if (EQ(cmd, "A_Reference")) {
		sercom_outint(arm_reference());
	} else goto unknown;

	arm_set_local(EQ(cmd, "A_RDis"));
	return true; 

table:
	if (EQ(cmd, "T_ID")) {
		char buf[MAXCMD];
		cerr = table_get_id(buf, sizeof(buf));
		sercom_outf("is", (uint32)cerr, buf);
	} else if (EQ(cmd, "T_Flags")) {
		cerr = table_get_flags(&flags);
		sercom_outf("ii", (uint32)cerr, flags);
	} else if (EQ(cmd, "T_RDis")) {
		/* TODO osetrit zakazani lok. ovladani */
		table_set_local(true);
		sercom_outint((uint32)cmd_ok);
	} else if (EQ(cmd, "T_Goto")) {
		int32 x, y;
		sercom_scanf(cmd, ".sII", &x, &y);
		sercom_outint(table_goto(x, y));
	} else if (EQ(cmd, "T_Kill")) {
		sercom_outint(table_kill());
	} else if (EQ(cmd, "T_Pos")) {
		int32 x, y;
		cerr = table_get_pos(&x, &y);
		sercom_outf("iII", (uint32)cerr, x, y);
	} else if (EQ(cmd, "T_SetZero")) {
		int32 x, y;
		sercom_scanf(cmd, ".sII", &x, &y);
		sercom_outint(table_set_zero(x, y));
	} else if (EQ(cmd, "T_GetZero")) {
		int32 x, y;
		cerr = table_get_zero(&x, &y);
		sercom_outf("iII", (uint32)cerr, x, y);
	} else if (EQ(cmd, "T_Reference")) {
		sercom_outint(table_go_reference());
	} else if (EQ(cmd, "T_Invert")) {
		uint32 invert;
		sercom_scanf(cmd, ".si", &invert);
		sercom_outint(table_invert(invert));
	/* Nedokumentovano (ale melo by): precte (a osetri) posledni chybu stolu */
	} else if (EQ(cmd, "T_LastErr")) { 
		uint32 last_error;
		cerr = table_get_last_error(&last_error);
		sercom_outf("ii", (uint32)cerr, last_error);
	/* Nedokumentovano: Posle stolu dotazovou zpravu a vrati odpoved */
	} else if (EQ(cmd, "T_Q")) { 
		char *query;
		char result[30];

		sercom_scanf(cmd, ".ss", &query);
		cerr = table_query(query, result, sizeof(result));
		sercom_outf("is", (uint32)cerr, result);
	} 
	return true;

unknown:
	puterror("PC", "?("); p_puts((const char *)cmd); p_putc(')');
	return true;
}

void
get_program(int program)
{
	struct arm_program *p = arm_programs + program;
	int row = 1;
		/* Je to hnusny, ale kdyz to nebude potreba predelavat.. proc ne. 
		 * (o 14 mesicu pozdeji: Samozrejme ze to predelavam, skoda ze neni
		 * videt ta puvodni hnusnost :-) */
	p_putscreen(
			"Program F\n"
			"Draha    xx.xx\001 \n"
			"\n"
			"Enter OK, ESC Zpet"
			, true);
	p_mvputc(8, 0, program + '2');
	p_move(8, 1); p_put_x_x(p->dist/10,  SPEC_X_X(X_X_SIGNED | X_X_ZEROS_AFTER, 2, 2));
	
	for (;;) {
		p_move(8, 1);
		p_getnum_x_x(NULL, SPEC_X_X(X_X_SIGNED, 2, 2));
		p_setcursor(C_NONE);
		switch(p_getkey()) {
			case K_ENTER: 
				p->dist = 10 * str2num_x_x(p_vram[1] + 8, SPEC_X_X(X_X_SIGNED, 2, 2));
				cmos_wr();
			case K_ESC:
				return;
			case K_NONE:
				break;
			default:
				buzzer_error();
		}
	}
}

static void 
help(void)
{
	p_putscreen(
			"MEAS Prog. s.r.o.\n"
			"www.measprog.com\n"
			"tel: 776 793 463\n"
			"Stiskni lib. klavesu", true);
	while(p_getkey() == K_NONE)
		sercmds_process();
}

/* Predpoklada, ze posledni funkce pro rameno vratila chybovy kod err. Pokud se
 * vratilo cmd_ok, vraci true. Pokud vratilo cmd_wait, vraci false. Pokud byla
 * vracena jina chyba, nevraci nic, vypise chybu a zasekne CKDM (uzivatel ji
 * musi resetovat). Pokud byla stisknuta havarijni stopka, celou situaci osetri
 * a vraci cmd_wait. Dale osetruje to, zda se rameno stale pohybuje, tj.  pokud
 * byla navratova hodnota cmd_wait, prepoklada ze rameno je stale v pohybu a
 * tudiz se musi menit odpor potenciometru pripojenemu k hrideli prevodovky.
 * Pokud se rameno nepohybuje, opet celou situaci nejak osetri.
 * Chyby, ktere mohou nastat, jsou nasledujici:
 * 1) Chyba komunikace s ramenem (cmd_invalid, cmd_timeout)
 * 2) Ztrata polohy (Sepnuti koncaku, neodpovidajici potak)
 * Chyby jsou setrizeny podle priority. 
 */ 
 
bool
handle_arm_cmd(enum cmd_err err)
{
	if (err == cmd_ok)
		return true;
	else if (err == cmd_wait)
		return false;
	puterror("Ram", cmd_strerr(err));
	return false;
}

/* chyby neplatneho pohybu ani nezobrazuji (uzivatel je na koncaku a macka
 * tlacitko ze chce jeste dal). Stejne uz to na displej pise 'Rameno mimo
 * rozsah'... */
enum cmd_err
handle_arm_go_cmd(enum cmd_err err)
{
	if (err == cmd_illegal_movement)
		return cmd_ok;
	return err;
}

/* Ceka az timeout jiffies nez rameno prestane byt busy (v pohybu). Vraci
 * cmd_ok pokud nakonec rameno neni busy, cmd_wait pokud je, pripadne jine
 * cmd_* pokud nastala chyba. */
enum cmd_err
busy_wait(int timeout)
{
	enum cmd_err err;
	uint32 timer, flags;

	timer = get_jiffies() + timeout; 
	do {
		if (cmd_ok != (err = arm_get_flags(&flags))) 
			return err;
		if (!(flags & ARM_BIT_BUSY)) 
			return cmd_ok;
		go_sleep();
	} while(get_jiffies() < timer);
	return cmd_wait;
}

static int32
round(int32 x, int32 step) 
{
	if (x > 0)
		return ((x + step/2)/step)*step;
	else
		return ((x - step/2)/step)*step;
}

/* Ovladani:
 * Nahoru/dolu: mikrokroky, drzeni: posun, drzeni se siftem: rychloposun
 * Shift-0 : Nastaveni nulove polohy do aktualni
 * Posta: Jed do nuly
 * Shift-posta: Referencuj
 * F1: Napoveda
 * F2..F5: Vyvolani programu 1 - 4
 * Shift-F2..Shift-F5: Uprava programu 1 - 4
 */

static void 
put_result_info(const char *device, enum cmd_err err)
{
	p_puts(device);
	p_putc(cmd_ok == err ? P_OK : P_BAD);
}


#ifdef TEST_INPUTS
static void
test_inputs(void) /* vypisuje AD1 .. AD7 + analog na AD0 */
{
	IO_ADER.byte = 0;
	IO_DDR6.bit.D60 = 0;
	IO_DDR6.bit.D61 = 0;
	IO_DDR6.bit.D62 = 0;
	IO_DDR6.bit.D63 = 0;
	IO_DDR6.bit.D64 = 0;
	IO_DDR6.bit.D65 = 0;
	IO_DDR6.bit.D66 = 0;
	IO_DDR6.bit.D67 = 0;
	ad_init();
	while(1) {
		p_msg();
	 	p_putc(IO_PDR6.bit.P61 ? '1': '0');
		p_putc(IO_PDR6.bit.P62 ? '1': '0');
		p_putc(IO_PDR6.bit.P63 ? '1': '0');
		p_putc(IO_PDR6.bit.P64 ? '1': '0');
		p_putc(IO_PDR6.bit.P65 ? '1': '0');
		p_putc(IO_PDR6.bit.P66 ? '1': '0');
		p_putc(IO_PDR6.bit.P67 ? '1': '0');
		p_putc(' ');
		p_putidf(ad_get_val(), 4); /* 0..1023 */
		jsleep(JIFRATE/3);
	}
}
#endif

void
ui(void)
{
	const byte p_deg[8] = {8, 20, 8, 0, 0, 0, 0, 0};
	const byte p_hourglass[8] = {31, 17, 14, 4, 4, 10, 31, 31};
	const byte p_ok[8] = {0, 0, 2, 2, 20, 20, 8, 0};
	/* typovaci vzdalenost [milistupne]. Na tuto vzdalenost je take
	 * zaokrouhlena hodnota na displeji. */
	const int type_distance = 50; 

	byte key;
	int32 mpos;
	uint32 tmp;
	enum arm_state arm_state;
	bool kill_sent = false;

	p_programchar(P_DEG, p_deg);
	p_programchar(P_HOURGLASS, p_hourglass);
	p_programchar(P_OK, p_ok);

#ifdef TEST_INPUTS
	test_inputs();
#endif

	p_clear();
	p_msg("Init komunikace s PC");
	sercmds_init();
	p_msg("Ctu nastaveni z CMOS");
	cmos_rd();
	p_msg("Init GPIB");
	gpib_init();
	gpib_ifc();
	p_msg("Init periferii");
	p_msg();
	p_puts("Rameno "); p_putc(cmd_ok == arm_init()   ? P_OK : P_BAD);
	p_puts(" Lux ");   p_putc(cmd_ok == lux_init()   ? P_OK : P_BAD);
	p_puts(" Stul ");  p_putc(cmd_ok == table_init() ? P_OK : P_BAD);

redraw:
	p_putscreen("\n\n\nF1: Vyrobce", true);

do_forever:

	if (sercmds_process())
		goto do_forever;	

	jsleep(JIFRATE/20);

	if (stop_button_pressed())
		goto do_forever;	

	key = p_getkey();

	if (!(arm_flags & ARM_BIT_ALIVE)) {
		p_clearrow(0); /* smazu radek se souradnicemi atd.. */
		p_mvputs(0, 1, "Rameno nenalezeno ");
		if (key == K_F1) {
			help();
			goto redraw;
		}
		if (key != K_NONE)
			arm_init();
		goto do_forever;
	}

	handle_arm_cmd(arm_get_pos(&mpos)); /* vraci wait kdyz rameno nestoji */
	handle_arm_cmd(arm_get_kill_sent(&kill_sent));
	handle_arm_cmd(arm_get_state(&arm_state));

	if (!kill_sent) {
		p_clearrow(0);
		p_move(0, 0);
		/* to /10 je tu kvuli prevodu mdeg (3 desetinna mista) na 2
		 * desetinna mista na displeji */
		p_put_x_x((round((mpos), TYPE_DISTANCE))/10,
			  SPEC_X_X(X_X_SIGNED | X_X_ZEROS_AFTER, 3, 2));
		p_mvputc(7, 0, P_DEG);
#ifdef SHOW_AD_VALUE
		p_move(0, 2);
		p_putidf(ad_get_val(), 4); /* 0..1023 */
#endif
	}

	handle_arm_cmd(arm_get_flags(&tmp));
	p_mvputc(19, 0, tmp & ARM_BIT_BUSY ? P_HOURGLASS : P_OK);
	p_mvputc(10, 0, tmp & ARM_BIT_REFERENCED ? ' ' : '?');
	p_mvputs(16, 0, tmp & ARM_BIT_LOCAL ? "  " : "PC");
	p_clearrow(1);
	if (tmp & ARM_BIT_RANGE)
		p_mvputs(0, 1, "Rameno mimo rozsah");

	if ((!(tmp & ARM_BIT_LOCAL)) && key != K_F1)
		goto do_forever;

	switch(key) {
		case K_F1:
			help();
			goto redraw;

		case K_UP:
			if (!p_autorepeating) {
				handle_arm_cmd(handle_arm_go_cmd(
					arm_goto(round((mpos + TYPE_DISTANCE), TYPE_DISTANCE))));
				break;
			}
			if (arm_state == go_positive || !handle_arm_cmd(busy_wait(JIFRATE/2)))
				break;
			handle_arm_cmd(arm_set_movement_params(P_ISSHIFT ? &arm_movement_full: &arm_movement_slow));
			handle_arm_cmd(handle_arm_go_cmd(arm_go_positive()));
			break;

		case K_DOWN:
			if (!p_autorepeating) {
				handle_arm_cmd(handle_arm_go_cmd(
					arm_goto(round((mpos - TYPE_DISTANCE), TYPE_DISTANCE))));
				break;
			}
			if (arm_state == go_negative || !handle_arm_cmd(busy_wait(JIFRATE/2)))
				break;
			handle_arm_cmd(arm_set_movement_params(P_ISSHIFT ? &arm_movement_full : &arm_movement_slow));
			handle_arm_cmd(handle_arm_go_cmd(arm_go_negative()));
			break;

		case K_0:
			if (P_ISSHIFT) {
				int32 zero;
				handle_arm_cmd(arm_get_zero(&zero)); 
				handle_arm_cmd(arm_set_zero(mpos + zero));
			}
			break;

		case K_BOX:
			handle_arm_cmd(arm_set_movement_params(&arm_movement_full));
			if (P_ISSHIFT) {
				handle_arm_cmd(arm_reference());
			} else {
				handle_arm_cmd(arm_goto(0));
			}
			break;

		case K_F2: case K_F3: case K_F4: case K_F5:
			tmp = key - K_F2;
			if (P_ISSHIFT && handle_arm_cmd(busy_wait(0))) {
				get_program(tmp);
				goto redraw;
			}
			handle_arm_cmd(arm_set_movement_params(&arm_movement_full));
			handle_arm_cmd(arm_goto(mpos + arm_programs[tmp].dist));
			break;
		case K_NONE:
			if (!p_autorepeating && (arm_state == go_positive || arm_state == go_negative)) {
				handle_arm_cmd(arm_kill());
			}
			break;
	} /* case */

	goto do_forever;
}
